/*
**                    TMM v1.0 for Avisynth 2.5.x
**
**   TMM builds a motion-mask for TDeint, which TDeint uses via
**   its 'emask' parameter.  TMM can use fixed or per-pixel adaptive
**   motion thresholds, as well as any length static period.  It
**   checks backwards, across, and forwards when looking for motion.
**
**   Copyright (C) 2007 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "buildMM.h"

buildMM::buildMM(PClip _tfields, PClip _bfields, int _mode, int _order,
	int _field, int _length, int _mtype, IScriptEnvironment *env) : 
	GenericVideoFilter(_tfields), bfields(_bfields), mode(_mode), order(_order), 
	field(_field), length(_length), mtype(_mtype)
{
	srcPFt = srcPFb = NULL;
	dstPF = NULL;
	gvlut = NULL;
	if (mode != 0 && mode != 1)
		env->ThrowError("TMM:  mode must be set to 0 or 1!");
	if (order < -1 || order > 1)
		env->ThrowError("TMM:  order must be set to -1, 0, or 1!");
	if (field < -1 || field > 1)
		env->ThrowError("TMM:  field must be set to -1, 0, or 1!");
	if (length < 6)
		env->ThrowError("TMM:  length must be greater than or equal to 6!");
	if (mtype != 0 && mtype != 1 && mtype != 2)
		env->ThrowError("TMM:  mtype must be set to 0, 1, or 2!");
	VideoInfo vi2 = bfields->GetVideoInfo();
	if (vi.height != vi2.height || vi.width != vi2.width ||
		vi.pixel_type != vi2.pixel_type || vi.num_frames != vi2.num_frames)
		env->ThrowError("TMM:  tfields and bfields clips don't match!");
	if (order == -1)
		order = child->GetParity(0) ? 1 : 0;
	if (field == -1)
		field = order;
	srcPFt = (PFCB**)calloc(length-2,sizeof(PFCB*));
	srcPFb = (PFCB**)calloc(length-2,sizeof(PFCB*));
	for (int i=0; i<length-2; ++i)
	{
		srcPFt[i] = (PFCB*)calloc(1,sizeof(PFCB));
		srcPFt[i]->pf = new PlanarFrame(vi);
		srcPFt[i]->n = -20;
		srcPFb[i] = (PFCB*)calloc(1,sizeof(PFCB));
		srcPFb[i]->pf = new PlanarFrame(vi);
		srcPFb[i]->n = -20;
	}
	vi_saved = vi;
	vi.height *= 2;
	if (mode == 1) vi.num_frames *= 2;
	dstPF = new PlanarFrame(vi);
	gvlut = (int*)malloc(length*sizeof(int));
	for (int i=0; i<length; ++i)
		gvlut[i] = i == 0 ? 1 : i == length-1 ? 4 : 2;
	vlut = mtype == 0 ? vlut2a : mtype == 1 ? vlut2b : vlut3;
}

buildMM::~buildMM()
{
	if (dstPF) delete dstPF;
	if (srcPFt)
	{
		for (int i=0; i<length-2; ++i)
		{
			if (srcPFt[i]) 
			{
				delete srcPFt[i]->pf;
				free(srcPFt[i]);
			}
		}
		free(srcPFt);
	}
	if (srcPFb)
	{
		for (int i=0; i<length-2; ++i)
		{
			if (srcPFb[i]) 
			{
				delete srcPFb[i]->pf;
				free(srcPFb[i]);
			}
		}
		free(srcPFb);
	}
	if (gvlut)
		free(gvlut);
}

void buildMM::resetFrames(PFCB **p, int start, int stop)
{
	for (int i=start; i<=stop; ++i)
	{
		for (int j=0; j<length-2; ++j)
		{
			if (p[j]->n == i && j != i-start)
			{
				PFCB *temp = p[i-start];
				p[i-start] = p[j];
				p[j] = temp;
				break;
			}
		}
	}
}

PVideoFrame __stdcall buildMM::GetFrame(int n, IScriptEnvironment *env)
{
	int fieldt = field;
	if (mode == 1)
	{
		fieldt = (n&1) ? 1-order : order;
		n >>= 1;
	}
	const int *tmmlut = tmmlut16[order]+(fieldt<<2);
	int tmmlutf[64];
	for (int i=0; i<64; ++i)
		tmmlutf[i] = tmmlut[vlut[i]];
	int ccount, ocount;
	int tstart, tstop, bstart, bstop;
	PFCB **cPF, **oPF;
	if (fieldt == 1)
	{
		tstart = n-((length-1)>>1);
		tstop = n+((length-1)>>1)-2;		
		const int bn = order == 1 ? n-1 : n;
		bstart = bn-((length-2)>>1);
		bstop = bn+1+((length-2)>>1)-2;
		ocount = tstop-tstart+1;
		ccount = bstop-bstart+1;
		oPF = srcPFt;
		cPF = srcPFb;
	}
	else
	{
		const int tn = order == 0 ? n-1 : n;
		tstart = tn-((length-2)>>1);
		tstop = tn+1+((length-2)>>1)-2;
		bstart = n-((length-1)>>1);
		bstop = n+((length-1)>>1)-2;
		ccount = tstop-tstart+1;
		ocount = bstop-bstart+1;
		cPF = srcPFt;
		oPF = srcPFb;
	}
	resetFrames(srcPFt,tstart,tstop);
	for (int i=tstart; i<=tstop; ++i)
	{
		if (srcPFt[i-tstart]->n == i)
			continue;
		srcPFt[i-tstart]->n = i;
		if (i < 0 || i >= vi_saved.num_frames-2)
			setBlack(srcPFt[i-tstart]->pf);
		else
			srcPFt[i-tstart]->pf->copyFrom(child->GetFrame(i,env),vi_saved);
	}
	resetFrames(srcPFb,bstart,bstop);
	for (int i=bstart; i<=bstop; ++i)
	{
		if (srcPFb[i-bstart]->n == i)
			continue;
		srcPFb[i-bstart]->n = i;
		if (i < 0 || i >= vi_saved.num_frames-2)
			setBlack(srcPFb[i-bstart]->pf);
		else
			srcPFb[i-bstart]->pf->copyFrom(bfields->GetFrame(i,env),vi_saved);
	}
	int *plut[2];
	for (int i=0; i<2; ++i)
		plut[i] = (int*)malloc((2*length-1)*sizeof(int));
	unsigned char **ptlut[3];
	for (int i=0; i<3; ++i)
		ptlut[i] = (unsigned char**)malloc(((i&1)?ccount:ocount)*sizeof(unsigned char*));
	const int offo = (length&1) ? 0 : 1;
	const int offc = (length&1) ? 1 : 0;
	for (int b=0; b<3; ++b)
	{
		unsigned char *dstp = dstPF->GetPtr(b);
		const int dst_pitch = dstPF->GetPitch(b);
		const int height = dstPF->GetHeight(b);
		const int width = dstPF->GetWidth(b);
		if (fieldt == 1)
		{
			for (int j=0; j<height; j+=2)
				memset(dstp+dst_pitch*j,10,width);
			dstp += dst_pitch;
		}
		else
		{
			for (int j=1; j<height; j+=2)
				memset(dstp+dst_pitch*j,10,width);
		}
		const int src_pitch = oPF[0]->pf->GetPitch(b);
		for (int i=0; i<ccount; ++i)
			ptlut[1][i] = cPF[i]->pf->GetPtr(b);
		for (int i=0; i<ocount; ++i)
		{
			if (fieldt == 1)
			{
				ptlut[0][i] = oPF[i]->pf->GetPtr(b);
				ptlut[2][i] = oPF[i]->pf->GetPtr(b)+src_pitch;
			}
			else
				ptlut[0][i] = ptlut[2][i] = oPF[i]->pf->GetPtr(b);
		}
		const int ct = ccount>>1;
		for (int y=fieldt; y<height; y+=2)
		{
			for (int x=0; x<width; ++x)
			{
				if (!ptlut[1][ct-2][x] && !ptlut[1][ct][x] &&
					!ptlut[1][ct+1][x])
				{
					dstp[x] = 60;
					continue;
				}
				for (int j=0; j<ccount; ++j)
					plut[0][j*2+offc] = plut[1][j*2+offc] = ptlut[1][j][x];
				for (int j=0; j<ocount; ++j)
				{
					plut[0][j*2+offo] = ptlut[0][j][x];
					plut[1][j*2+offo] = ptlut[2][j][x];
				}
				int val = 0;
				for (int i=0; i<length; ++i)
				{
					for (int j=0; j<length-4; ++j)
					{
						if (!plut[0][i+j])
							goto j1;
					}
					val |= (gvlut[i]<<3);
			j1:
					for (int j=0; j<length-4; ++j)
					{
						if (!plut[1][i+j])
							goto j2;
					}
					val |= gvlut[i];
			j2:
					if (vlut[val] == 2)
						break;
				}
				dstp[x] = tmmlutf[val];
			}
			dstp += dst_pitch*2;
			for (int i=0; i<ccount; ++i)
				ptlut[1][i] += src_pitch;
			for (int i=0; i<ocount; ++i)
			{
				if (y != 0) ptlut[0][i] += src_pitch;
				if (y != height-3) ptlut[2][i] += src_pitch;
			}
		}
	}
	for (int i=0; i<3; ++i)
		free(ptlut[i]);
	for (int i=0; i<2; ++i)
		free(plut[i]);
	PVideoFrame dst = env->NewVideoFrame(vi);
	dstPF->copyTo(dst,vi);
	return dst;
}

void buildMM::setBlack(PlanarFrame *p)
{
	for (int b=0; b<3; ++b)
		memset(p->GetPtr(b),0,p->GetPitch(b)*p->GetHeight(b));
}