/*
**                    TMM v1.0 for Avisynth 2.5.x
**
**   TMM builds a motion-mask for TDeint, which TDeint uses via
**   its 'emask' parameter.  TMM can use fixed or per-pixel adaptive
**   motion thresholds, as well as any length static period.  It
**   checks backwards, across, and forwards when looking for motion.
**
**   Copyright (C) 2007 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "createMM.h"

createMM::createMM(PClip _child, int _ttype, int _mtqL, int _mthL, int _mtqC, 
	int _mthC, int _nt, int _minthresh, int _maxthresh, int _cstr, 
	IScriptEnvironment *env) : GenericVideoFilter(_child), ttype(_ttype), 
	mtqL(_mtqL), mthL(_mthL), mtqC(_mtqC), mthC(_mthC), nt(_nt), minthresh(_minthresh), 
	maxthresh(_maxthresh), cstr(_cstr)
{
	frames = NULL;
	for (int i=0; i<2; ++i)
		dstPF[i] = NULL;
	for (int i=0; i<3; ++i)
		offplut[i] = offnlut[i] = NULL;
	if (!vi.IsYV12() && !vi.IsYUY2())
		env->ThrowError("TMM:  only yv12 and yuy2 input supported!");
	if (ttype < 0 || ttype > 5)
		env->ThrowError("TMM:  ttype must be set to 0, 1, 2, 3, 4, or 5!");
	if (mtqL < -1 || mtqL > 255)
		env->ThrowError("TMM:  mtqL must be between -1 and 255 inclusive!");
	if (mtqC < -1 || mtqC > 255)
		env->ThrowError("TMM:  mtqC must be between -1 and 255 inclusive!");
	if (mthL < -1 || mthL > 255)
		env->ThrowError("TMM:  mthL must be between -1 and 255 inclusive!");
	if (mthC < -1 || mthC > 255)
		env->ThrowError("TMM:  mthC must be between -1 and 255 inclusive!");
	frames = (PFCC**)calloc(3,sizeof(PFCC*));
	for (int i=0; i<3; ++i)
	{
		frames[i] = (PFCC*)calloc(1,sizeof(PFCC));
		frames[i]->pf[0] = new PlanarFrame(vi);
		vi.height *= 2;
		frames[i]->pf[1] = new PlanarFrame(vi);
		frames[i]->pf[2] = new PlanarFrame(vi);
		vi.height /= 2;
		frames[i]->n[0] = frames[i]->n[1] = -20;
	}
	vi.height *= 2;
	dstPF[0] = new PlanarFrame(vi);
	vi.height /= 2;
	dstPF[1] = new PlanarFrame(vi);
	for (int i=0; i<3; ++i)
	{
		const int width = dstPF[0]->GetWidth(i);
		offplut[i] = (int*)malloc(width*sizeof(int));
		offnlut[i] = (int*)malloc(width*sizeof(int));
		for (int j=0; j<width; ++j)
		{
			if (j == 0) offplut[i][j] = -1;
			else offplut[i][j] = 1;
			if (j == width-1) offnlut[i][j] = -1;
			else offnlut[i][j] = 1;
		}
	}
	for (int i=0; i<256; ++i)
		mlut[i] = min(max(i+nt,minthresh),maxthresh);
}

createMM::~createMM()
{
	for (int i=0; i<3; ++i)
	{
		if (offnlut[i]) free(offnlut[i]);
		if (offplut[i]) free(offplut[i]);
	}
	for (int i=0; i<2; ++i)
	{
		if (dstPF[i]) 
			delete dstPF[i];
	}
	if (frames)
	{
		for (int i=0; i<3; ++i)
		{
			if (frames[i])
			{
				for (int j=0; j<3; ++j)
				{
					if (frames[i]->pf[j])
						delete frames[i]->pf[j];
				}
				free(frames[i]);
			}
		}
		free(frames);
	}
}

void createMM::resetFrames(PFCC **p, int start, int stop)
{
	for (int i=start; i<=stop; ++i)
	{
		for (int j=0; j<3; ++j)
		{
			if (p[j]->n[0] == i && j != i-start)
			{
				PFCC *temp = p[i-start];
				p[i-start] = p[j];
				p[j] = temp;
				break;
			}
		}
	}
}

int createMM::mapn(int n)
{
	if (n < 0) return 0;
	if (n > vi.num_frames-1) return vi.num_frames-1;
	return n;
}

PVideoFrame __stdcall createMM::GetFrame(int n, IScriptEnvironment *env)
{
	resetFrames(frames,n,n+2);
	for (int i=n; i<=n+2; ++i)
	{
		if (frames[i-n]->n[0] != i)
		{
			frames[i-n]->pf[0]->copyFrom(child->GetFrame(mapn(i),env),vi);
			threshMask(frames[i-n]->pf[0],frames[i-n]->pf[1]);
			frames[i-n]->n[0] = i;
		}
	}
	for (int i=n; i<=n+1; ++i)
	{
		if (frames[i-n]->n[1] != i)
		{
			motionMask(frames[i-n]->pf[0],frames[i-n]->pf[1],
				frames[i-n+1]->pf[0],frames[i-n+1]->pf[1],
				frames[i-n]->pf[2]);
			frames[i-n]->n[1] = i;
		}
	}
	motionMask(frames[0]->pf[0],frames[0]->pf[1],
		frames[2]->pf[0],frames[2]->pf[1],dstPF[0]);
	andMasks(dstPF[0],frames[0]->pf[2],frames[1]->pf[2]);
	combineMasks(dstPF[0],dstPF[1],env);
	PVideoFrame dst = env->NewVideoFrame(vi);
	dstPF[1]->copyTo(dst,vi);
	return dst;
}

void createMM::combineMasks(PlanarFrame *src, PlanarFrame *dst, 
	IScriptEnvironment *env)
{
	for (int b=0; b<3; ++b)
	{
		const unsigned char *srcp0 = src->GetPtr(b);
		const int pitch = src->GetPitch(b);
		const int width = src->GetWidth(b);
		const int height = dst->GetHeight(b);
		unsigned char *dstp = dst->GetPtr(b);
		env->BitBlt(dstp,pitch,srcp0,pitch,width,height);
		const unsigned char *srcp1 = srcp0+height*pitch;
		const int *offpt = offplut[b];
		const int *offnt = offnlut[b];
		for (int y=0; y<height; ++y)
		{
			const unsigned char *srcpp0 = srcp0-(y==0?-pitch:pitch);
			const unsigned char *srcpn0 = srcp0+(y==height-1?-pitch:pitch);
			for (int x=0; x<width; ++x)
			{
				if (srcp0[x] || !srcp1[x])
					continue;
				const int offp = offpt[x];
				const int offn = offnt[x];
				int count = 0;
				if (srcpp0[x-offp]) ++count;
				if (srcpp0[x]) ++count;
				if (srcpp0[x+offn]) ++count;
				if (srcp0[x-offp]) ++count;
				if (srcp0[x+offn]) ++count;
				if (srcpn0[x-offp]) ++count;
				if (srcpn0[x]) ++count;
				if (srcpn0[x+offn]) ++count;
				if (count >= cstr)
					dstp[x] = 255;
			}
			srcp0 += pitch;
			srcp1 += pitch;
			dstp += pitch;
		}
	}
}

void createMM::andMasks(PlanarFrame *dst, PlanarFrame *src1, 
		PlanarFrame *src2)
{
	for (int b=0; b<3; ++b)
	{
		const unsigned char *srcp1 = src1->GetPtr(b);
		const unsigned char *srcp2 = src2->GetPtr(b);
		const int pitch = src1->GetPitch(b);
		const int height = src1->GetHeight(b);
		const int width = src1->GetWidth(b);
		unsigned char *dstp = dst->GetPtr(b);
		for (int y=0; y<height; ++y)
		{
			for (int x=0; x<width; ++x)
				dstp[x] &= (srcp1[x]&srcp2[x]);
			dstp += pitch;
			srcp1 += pitch;
			srcp2 += pitch;
		}
	}
}

void createMM::motionMask(PlanarFrame *src1, PlanarFrame *msk1,
	PlanarFrame *src2, PlanarFrame *msk2, PlanarFrame *dst)
{
	for (int b=0; b<3; ++b)
	{
		const int pitch = src1->GetPitch(b);
		const int width = src1->GetWidth(b);
		const int height = src1->GetHeight(b);
		const unsigned char *srcp1 = src1->GetPtr(b);
		const unsigned char *srcp2 = src2->GetPtr(b);
		const unsigned char *mskp1q = msk1->GetPtr(b);
		const unsigned char *mskp1h = mskp1q+height*pitch;
		const unsigned char *mskp2q = msk2->GetPtr(b);
		const unsigned char *mskp2h = mskp2q+height*pitch;
		unsigned char *dstpq = dst->GetPtr(b);
		unsigned char *dstph = dstpq+height*pitch;
		for (int y=0; y<height; ++y)
		{
			for (int x=0; x<width; ++x)
			{
				const int diff = abs(srcp1[x]-srcp2[x]);
				const int threshq = min(mskp1q[x],mskp2q[x]);
				if (diff <= mlut[threshq])
					dstpq[x] = 255;
				else dstpq[x] = 0;
				const int threshh = min(mskp1h[x],mskp2h[x]);
				if (diff <= mlut[threshh])
					dstph[x] = 255;
				else dstph[x] = 0;
			}
			srcp1 += pitch;
			srcp2 += pitch;
			mskp1q += pitch;
			mskp1h += pitch;
			mskp2q += pitch;
			mskp2h += pitch;
			dstpq += pitch;
			dstph += pitch;
		}
	}
}

void createMM::threshMask(PlanarFrame *src, PlanarFrame *dst)
{
	for (int b=0; b<3; ++b)
	{
		const unsigned char *srcp = src->GetPtr(b);
		const int pitch = src->GetPitch(b);
		const int width = src->GetWidth(b);
		const int height = src->GetHeight(b);
		unsigned char *dstp0 = dst->GetPtr(b);
		unsigned char *dstp1 = dstp0+height*pitch;
		if (b == 0 && mtqL > -1 && mthL > -1)
		{
			memset(dstp0,mtqL,height*pitch);
			memset(dstp1,mthL,height*pitch);
			continue;
		}
		else if (b > 0 && mtqC > -1 && mthC > -1)
		{
			memset(dstp0,mtqC,height*pitch);
			memset(dstp1,mthC,height*pitch);
			continue;
		}
		const int hs = b == 0 ? 0 : 1;
		const int vs = b == 0 ? 1 : (vi.IsYV12() ? 2 : 1);
		const int *offpt = offplut[b];
		const int *offnt = offnlut[b];
		if (ttype == 0) // 4 neighbors - compensated
		{
			for (int y=0; y<height; ++y)
			{
				const unsigned char *srcpp = srcp-(y==0?-pitch:pitch);
				const unsigned char *srcpn = srcp+(y==height-1?-pitch:pitch);
				for (int x=0; x<width; ++x)
				{
					const int offp = offpt[x];
					const int offn = offnt[x];
					int min0 = 256, max0 = -1;
					int min1 = 256, max1 = -1;
					if (srcpp[x] < min0) min0 = srcpp[x];
					if (srcpp[x] > max0) max0 = srcpp[x];
					if (srcp[x-offp] < min1) min1 = srcp[x-offp];
					if (srcp[x-offp] > max1) max1 = srcp[x-offp];
					if (srcp[x+offn] < min1) min1 = srcp[x+offn];
					if (srcp[x+offn] > max1) max1 = srcp[x+offn];
					if (srcpn[x] < min0) min0 = srcpn[x];
					if (srcpn[x] > max0) max0 = srcpn[x];
					const int atv = max((abs(srcp[x]-min0)+vs)>>vs,(abs(srcp[x]-max0)+vs)>>vs);
					const int ath = max((abs(srcp[x]-min1)+hs)>>hs,(abs(srcp[x]-max1)+hs)>>hs);
					dstp0[x] = (max(atv,ath)+2)>>2;
					dstp1[x] = (max(atv,ath)+1)>>1;
				}
				srcp += pitch;
				dstp0 += pitch;
				dstp1 += pitch;
			}
		}
		else if (ttype == 1) // 8 neighbors - compensated
		{
			for (int y=0; y<height; ++y)
			{
				const unsigned char *srcpp = srcp-(y==0?-pitch:pitch);
				const unsigned char *srcpn = srcp+(y==height-1?-pitch:pitch);
				for (int x=0; x<width; ++x)
				{
					const int offp = offpt[x];
					const int offn = offnt[x];
					int min0 = 256, max0 = -1;
					int min1 = 256, max1 = -1;
					if (srcpp[x-offp] < min0) min0 = srcpp[x-offp];
					if (srcpp[x-offp] > max0) max0 = srcpp[x-offp];
					if (srcpp[x] < min0) min0 = srcpp[x];
					if (srcpp[x] > max0) max0 = srcpp[x];
					if (srcpp[x+offn] < min0) min0 = srcpp[x+offn];
					if (srcpp[x+offn] > max0) max0 = srcpp[x+offn];
					if (srcp[x-offp] < min1) min1 = srcp[x-offp];
					if (srcp[x-offp] > max1) max1 = srcp[x-offp];
					if (srcp[x+offn] < min1) min1 = srcp[x+offn];
					if (srcp[x+offn] > max1) max1 = srcp[x+offn];
					if (srcpn[x-offp] < min0) min0 = srcpn[x-offp];
					if (srcpn[x-offp] > max0) max0 = srcpn[x-offp];
					if (srcpn[x] < min0) min0 = srcpn[x];
					if (srcpn[x] > max0) max0 = srcpn[x];
					if (srcpn[x+offn] < min0) min0 = srcpn[x+offn];
					if (srcpn[x+offn] > max0) max0 = srcpn[x+offn];
					const int atv = max((abs(srcp[x]-min0)+vs)>>vs,(abs(srcp[x]-max0)+vs)>>vs);
					const int ath = max((abs(srcp[x]-min1)+hs)>>hs,(abs(srcp[x]-max1)+hs)>>hs);
					dstp0[x] = (max(atv,ath)+2)>>2;
					dstp1[x] = (max(atv,ath)+1)>>1;
				}
				srcp += pitch;
				dstp0 += pitch;
				dstp1 += pitch;
			}
		}
		else if (ttype == 2) // 4 neighbors - not compensated
		{
			for (int y=0; y<height; ++y)
			{
				const unsigned char *srcpp = srcp-(y==0?-pitch:pitch);
				const unsigned char *srcpn = srcp+(y==height-1?-pitch:pitch);
				for (int x=0; x<width; ++x)
				{
					const int offp = offpt[x];
					const int offn = offnt[x];
					int min0 = 256, max0 = -1;
					if (srcpp[x] < min0) min0 = srcpp[x];
					if (srcpp[x] > max0) max0 = srcpp[x];
					if (srcp[x-offp] < min0) min0 = srcp[x-offp];
					if (srcp[x-offp] > max0) max0 = srcp[x-offp];
					if (srcp[x+offn] < min0) min0 = srcp[x+offn];
					if (srcp[x+offn] > max0) max0 = srcp[x+offn];
					if (srcpn[x] < min0) min0 = srcpn[x];
					if (srcpn[x] > max0) max0 = srcpn[x];
					const int at = max(abs(srcp[x]-min0),abs(srcp[x]-max0));
					dstp0[x] = (at+2)>>2;
					dstp1[x] = (at+1)>>1;
				}
				srcp += pitch;
				dstp0 += pitch;
				dstp1 += pitch;
			}
		}
		else if (ttype == 3) // 8 neighbors - not compensated
		{
			for (int y=0; y<height; ++y)
			{
				const unsigned char *srcpp = srcp-(y==0?-pitch:pitch);
				const unsigned char *srcpn = srcp+(y==height-1?-pitch:pitch);
				for (int x=0; x<width; ++x)
				{
					const int offp = offpt[x];
					const int offn = offnt[x];
					int min0 = 256, max0 = -1;
					if (srcpp[x-offp] < min0) min0 = srcpp[x-offp];
					if (srcpp[x-offp] > max0) max0 = srcpp[x-offp];
					if (srcpp[x] < min0) min0 = srcpp[x];
					if (srcpp[x] > max0) max0 = srcpp[x];
					if (srcpp[x+offn] < min0) min0 = srcpp[x+offn];
					if (srcpp[x+offn] > max0) max0 = srcpp[x+offn];
					if (srcp[x-offp] < min0) min0 = srcp[x-offp];
					if (srcp[x-offp] > max0) max0 = srcp[x-offp];
					if (srcp[x+offn] < min0) min0 = srcp[x+offn];
					if (srcp[x+offn] > max0) max0 = srcp[x+offn];
					if (srcpn[x-offp] < min0) min0 = srcpn[x-offp];
					if (srcpn[x-offp] > max0) max0 = srcpn[x-offp];
					if (srcpn[x] < min0) min0 = srcpn[x];
					if (srcpn[x] > max0) max0 = srcpn[x];
					if (srcpn[x+offn] < min0) min0 = srcpn[x+offn];
					if (srcpn[x+offn] > max0) max0 = srcpn[x+offn];
					const int at = max(abs(srcp[x]-min0),abs(srcp[x]-max0));
					dstp0[x] = (at+2)>>2;
					dstp1[x] = (at+1)>>1;
				}
				srcp += pitch;
				dstp0 += pitch;
				dstp1 += pitch;
			}
		}
		else if (ttype == 4) // 4 neighbors - not compensated (range)
		{
			for (int y=0; y<height; ++y)
			{
				const unsigned char *srcpp = srcp-(y==0?-pitch:pitch);
				const unsigned char *srcpn = srcp+(y==height-1?-pitch:pitch);
				for (int x=0; x<width; ++x)
				{
					const int offp = offpt[x];
					const int offn = offnt[x];
					int min0 = 256, max0 = -1;
					if (srcpp[x] < min0) min0 = srcpp[x];
					if (srcpp[x] > max0) max0 = srcpp[x];
					if (srcp[x-offp] < min0) min0 = srcp[x-offp];
					if (srcp[x-offp] > max0) max0 = srcp[x-offp];
					if (srcp[x] < min0) min0 = srcp[x];
					if (srcp[x] > max0) max0 = srcp[x];
					if (srcp[x+offn] < min0) min0 = srcp[x+offn];
					if (srcp[x+offn] > max0) max0 = srcp[x+offn];
					if (srcpn[x] < min0) min0 = srcpn[x];
					if (srcpn[x] > max0) max0 = srcpn[x];
					const int at = max0-min0;
					dstp0[x] = (at+2)>>2;
					dstp1[x] = (at+1)>>1;
				}
				srcp += pitch;
				dstp0 += pitch;
				dstp1 += pitch;
			}
		}
		else if (ttype == 5) // 8 neighbors - not compensated (range)
		{
			for (int y=0; y<height; ++y)
			{
				const unsigned char *srcpp = srcp-(y==0?-pitch:pitch);
				const unsigned char *srcpn = srcp+(y==height-1?-pitch:pitch);
				for (int x=0; x<width; ++x)
				{
					const int offp = offpt[x];
					const int offn = offnt[x];
					int min0 = 256, max0 = -1;
					if (srcpp[x-offp] < min0) min0 = srcpp[x-offp];
					if (srcpp[x-offp] > max0) max0 = srcpp[x-offp];
					if (srcpp[x] < min0) min0 = srcpp[x];
					if (srcpp[x] > max0) max0 = srcpp[x];
					if (srcpp[x+offn] < min0) min0 = srcpp[x+offn];
					if (srcpp[x+offn] > max0) max0 = srcpp[x+offn];
					if (srcp[x-offp] < min0) min0 = srcp[x-offp];
					if (srcp[x-offp] > max0) max0 = srcp[x-offp];
					if (srcp[x] < min0) min0 = srcp[x];
					if (srcp[x] > max0) max0 = srcp[x];
					if (srcp[x+offn] < min0) min0 = srcp[x+offn];
					if (srcp[x+offn] > max0) max0 = srcp[x+offn];
					if (srcpn[x-offp] < min0) min0 = srcpn[x-offp];
					if (srcpn[x-offp] > max0) max0 = srcpn[x-offp];
					if (srcpn[x] < min0) min0 = srcpn[x];
					if (srcpn[x] > max0) max0 = srcpn[x];
					if (srcpn[x+offn] < min0) min0 = srcpn[x+offn];
					if (srcpn[x+offn] > max0) max0 = srcpn[x+offn];
					const int at = max0-min0;
					dstp0[x] = (at+2)>>2;
					dstp1[x] = (at+1)>>1;
				}
				srcp += pitch;
				dstp0 += pitch;
				dstp1 += pitch;
			}
		}
		if (b == 0 && mtqL > -1)
			memset(dst->GetPtr(b),mtqL,height*pitch);
		else if (b == 0 && mthL > -1)
			memset(dst->GetPtr(b)+height*pitch,mthL,height*pitch);
		else if (b > 0 && mtqC > -1)
			memset(dst->GetPtr(b),mtqC,height*pitch);
		else if (b > 0 && mthC > -1)
			memset(dst->GetPtr(b)+height*pitch,mthC,height*pitch);
	}
}